#!/usr/bin/env python3
from __future__ import print_function

# fillchk.py: scan for repetitions of a single byte (file corruption)
# 2023-11-26,  ed <fillchk@ocv.me>,  MIT-licensed
#
# this impl on files with 0% corruption:
#  1242 MiB/s in python3
#  1247 MiB/s in python2
#  4030 MiB/s in pypy2
#  1140 MiB/s in pypy39
#
# this impl on files with 100% corruption:
#     9 MiB/s in python3
#    10 MiB/s in python2
#   295 MiB/s in pypy2
#   505 MiB/s in pypy39
#
# see also:
#   https://ocv.me/dev/fillchk.c
#   https://ocv.me/dev/fillchk.cpp
#   https://ocv.me/dev/fillchk.py
#   https://ocv.me/dev/fillchk.pl
#   https://ocv.me/dev/fillchk.go
#   https://ocv.me/dev/fillchk.lua
#
# new in v1.3:
# - huge speedup in regions with no repetitions/corruption

import os
import sys
import time
import struct
import platform

MAX_REPS = 0x400  # number of repetitions before a corrupted range is assumed

PY2 = sys.version_info[0] == 2

def msg_filler(ch, p1, p2):
	ch = struct.unpack(b'B', ch)[0] if PY2 else ch
	t = "from %08x to %08x, %02x seen 0x%x / %s times\x1b[J"
	print(t % (p1, p2, ch, p2-p1, p2-p1))

def msg_speed(t1, t2, p1, p2, suffix=""):
	t = "at %08x, %.2f MiB/s%s"
	t = t % (p2, ((p2 - p1) / (t2 - t1)) / (1024*1024.0), suffix)
	print(t, file=sys.stderr)

def main():
	if len(sys.argv) > 2:
		print('need either...', file=sys.stderr)
		print('  0 args (will read from stdin)', file=sys.stderr)
		print('  1 args (filepath to read from)', file=sys.stderr)
		sys.exit(1)

	elif len(sys.argv) == 2:
		print('using filepath', file=sys.stderr)
		read_sz = 64 * 1024
		fh = open(sys.argv[1], 'rb', read_sz)

	else:
		print('using stdin', file=sys.stderr)
		read_sz = 16 * 1024
		if not PY2:
			fh = sys.stdin.buffer
		else:
			fh = sys.stdin
			if sys.platform == 'win32':
				if platform.python_implementation() == "PyPy":
					print('workaround: win32-pypy2', file=sys.stderr)
					fh = os.fdopen(fh.fileno(), "rb", 0)
				else:
					print('workaround: win32-cpy2', file=sys.stderr)
					import msvcrt
					msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)

	bad_segs = 0
	pos = 0
	n_dupes = 1
	dupe_chr = -1
	t0 = time.time()
	prev_time = t0
	prev_pos = 0
	while True:
		buf = fh.read(read_sz)
		if not buf:
			break

		buf_ctr = 0
		i = 0
		imax = len(buf)
		while i < imax:
			ch = buf[i]
			buf_ctr += 1
			if ch == dupe_chr:
				n_dupes += 1
				i += 1
				continue

			if n_dupes >= MAX_REPS:
				pos += buf_ctr
				buf_ctr = 0
				msg_filler(dupe_chr, pos - n_dupes - 1, pos - 1)
				bad_segs += 1

			# skip the next MAX_REPS-3 bytes if current byte differs from the one at that offset, as there are no repetitions longer than MAX_REPS in this range
			i2 = i + MAX_REPS - 3
			if i2 < imax and ch != buf[i2]:
				buf_ctr += i2 - i
				i = i2
				ch = buf[i]
				while ch == buf[i-1]:
					buf_ctr -= 1
					i -= 1

			i += 1
			n_dupes = 1
			dupe_chr = ch

		pos += buf_ctr
		now = time.time()
		if now - prev_time >= 1:
			msg_speed(prev_time, now, prev_pos, pos)
			prev_time = now
			prev_pos = pos

	if n_dupes >= MAX_REPS:
		msg_filler(dupe_chr, pos - n_dupes, pos)
		bad_segs += 1

	msg_speed(t0, time.time(), 0, pos, " (total avg)")

	sys.exit(min(255, bad_segs))

if __name__ == '__main__':
	main()
